/* wrap commands instead of scrolling horizontally */
.show_and_tell_command > pre > code {
    white-space: normal !important;
}

/* center captions */
.md-typeset .caption,
.md-typeset figcaption {
    /* same as material's figcaption */
    font-style: normal;
    max-width: 24rem;
    margin: 1em auto 1.5em;
    text-align: center;
    display: block;
    font-size: 0.8rem;
    /* custom */
    line-height: 1.2;
}
